@@ -1,5 +1,5 @@
-/*  Clzip - Data compressor based on the LZMA algorithm
-    Copyright (C) 2010, 2011 Antonio Diaz Diaz.
+/*  Clzip - LZMA lossless data compressor
+    Copyright (C) 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 /*
-    Return values: 0 for a normal exit, 1 for environmental problems
+    Exit status: 0 for a normal exit, 1 for environmental problems
     (file not found, invalid flags, I/O errors, etc), 2 to indicate a
     corrupt or invalid input file, 3 for an internal consistency error
     (eg, bug) which caused clzip to panic.
@@ -24,14 +24,14 @@
 #define _FILE_OFFSET_BITS 64
 
 #include <errno.h>
+#include <fcntl.h>
 #include <limits.h>
 #include <signal.h>
-#include <stdio.h>
 #include <stdbool.h>
+#include <stdint.h>
+#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <fcntl.h>
-#include <stdint.h>
 #include <unistd.h>
 #include <utime.h>
 #include <sys/stat.h>
@@ -39,6 +39,7 @@
 #include <io.h>
 #define fchmod(x,y) 0
 #define fchown(x,y,z) 0
+#define strtoull strtoul
 #define SIGHUP SIGTERM
 #define S_ISSOCK(x) 0
 #define S_IRGRP 0
@@ -51,7 +52,7 @@
 #endif
 
 #include "carg_parser.h"
-#include "clzip.h"
+#include "lzip.h"
 #include "decoder.h"
 #include "encoder.h"
 
@@ -59,22 +60,10 @@
 #error "Environments where CHAR_BIT != 8 are not supported."
 #endif
 
-#ifndef LLONG_MAX
-#define LLONG_MAX  0x7FFFFFFFFFFFFFFFLL
-#endif
-#ifndef LLONG_MIN
-#define LLONG_MIN  (-LLONG_MAX - 1LL)
-#endif
-#ifndef ULLONG_MAX
-#define ULLONG_MAX 0xFFFFFFFFFFFFFFFFULL
-#endif
-
-long long int llabs( long long int number );
-
 
 const char * const Program_name = "Clzip";
 const char * const program_name = "clzip";
-const char * const program_year = "2011";
+const char * const program_year = "2013";
 const char * invocation_name = 0;
 
 #ifdef O_BINARY
@@ -90,8 +79,8 @@ struct { const char * from; const char * to; } const known_extensions[] = {
 
 struct Lzma_options
   {
-  int dictionary_size;		/* 4KiB..512MiB */
-  int match_len_limit;		/* 5..273 */
+  int dictionary_size;		/* 4 KiB .. 512 MiB */
+  int match_len_limit;		/* 5 .. 273 */
   };
 
 enum Mode { m_compress, m_decompress, m_test };
@@ -99,90 +88,88 @@ enum Mode { m_compress, m_decompress, m_test };
 char * output_filename = 0;
 int outfd = -1;
 int verbosity = 0;
+const mode_t usr_rw = S_IRUSR | S_IWUSR;
+const mode_t all_rw = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
 mode_t outfd_mode = S_IRUSR | S_IWUSR;
 bool delete_output_on_interrupt = false;
 
 
-/* assure at least a minimum size for buffer `buf' */
-static void * resize_buffer( void * buf, const int min_size )
+static void show_help( void )
   {
-  if( buf ) buf = realloc( buf, min_size );
-  else buf = malloc( min_size );
-  if( !buf )
-    {
-    show_error( "Not enough memory.", 0, false );
-    cleanup_and_fail( 1 );
-    }
-  return buf;
-  }
-
-
-static void show_help()
-  {
-  printf( "%s - Data compressor based on the LZMA algorithm.\n", Program_name );
+  printf( "%s - LZMA lossless data compressor.\n", Program_name );
   printf( "\nUsage: %s [options] [files]\n", invocation_name );
-  printf( "\nOptions:\n" );
-  printf( "  -h, --help                 display this help and exit\n" );
-  printf( "  -V, --version              output version information and exit\n" );
-  printf( "  -b, --member-size=<n>      set member size limit in bytes\n" );
-  printf( "  -c, --stdout               send output to standard output\n" );
-  printf( "  -d, --decompress           decompress\n" );
-  printf( "  -f, --force                overwrite existing output files\n" );
-  printf( "  -F, --recompress           force recompression of compressed files\n" );
-  printf( "  -k, --keep                 keep (don't delete) input files\n" );
-  printf( "  -m, --match-length=<n>     set match length limit in bytes [36]\n" );
-  printf( "  -o, --output=<file>        if reading stdin, place the output into <file>\n" );
-  printf( "  -q, --quiet                suppress all messages\n" );
-  printf( "  -s, --dictionary-size=<n>  set dictionary size limit in bytes [8MiB]\n" );
-  printf( "  -S, --volume-size=<n>      set volume size limit in bytes\n" );
-  printf( "  -t, --test                 test compressed file integrity\n" );
-  printf( "  -v, --verbose              be verbose (a 2nd -v gives more)\n" );
-  printf( "  -1 .. -9                   set compression level [default 6]\n" );
-  printf( "      --fast                 alias for -1\n" );
-  printf( "      --best                 alias for -9\n" );
-  printf( "If no file names are given, %s compresses or decompresses\n", program_name );
-  printf( "from standard input to standard output.\n" );
-  printf( "Numbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n" );
-  printf( "Ki = KiB = 2^10 = 1024, M = 10^6, Mi = 2^20, G = 10^9, Gi = 2^30, etc...\n" );
-  printf( "\nReport bugs to lzip-bug@nongnu.org\n" );
-  printf( "Clzip home page: http://www.nongnu.org/lzip/clzip.html\n" );
+  printf( "\nOptions:\n"
+          "  -h, --help                     display this help and exit\n"
+          "  -V, --version                  output version information and exit\n"
+          "  -b, --member-size=<bytes>      set member size limit in bytes\n"
+          "  -c, --stdout                   send output to standard output\n"
+          "  -d, --decompress               decompress\n"
+          "  -f, --force                    overwrite existing output files\n"
+          "  -F, --recompress               force recompression of compressed files\n"
+          "  -k, --keep                     keep (don't delete) input files\n"
+          "  -m, --match-length=<bytes>     set match length limit in bytes [36]\n"
+          "  -o, --output=<file>            if reading stdin, place the output into <file>\n"
+          "  -q, --quiet                    suppress all messages\n"
+          "  -s, --dictionary-size=<bytes>  set dictionary size limit in bytes [8 MiB]\n"
+          "  -S, --volume-size=<bytes>      set volume size limit in bytes\n"
+          "  -t, --test                     test compressed file integrity\n"
+          "  -v, --verbose                  be verbose (a 2nd -v gives more)\n"
+          "  -1 .. -9                       set compression level [default 6]\n"
+          "      --fast                     alias for -1\n"
+          "      --best                     alias for -9\n"
+          "If no file names are given, clzip compresses or decompresses\n"
+          "from standard input to standard output.\n"
+          "Numbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n"
+          "Ki = KiB = 2^10 = 1024, M = 10^6, Mi = 2^20, G = 10^9, Gi = 2^30, etc...\n"
+          "The bidimensional parameter space of LZMA can't be mapped to a linear\n"
+          "scale optimal for all files. If your files are large, very repetitive,\n"
+          "etc, you may need to use the --match-length and --dictionary-size\n"
+          "options directly to achieve optimal performance. For example, -9m64\n"
+          "usually compresses executables more (and faster) than -9.\n"
+          "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
+          "not found, invalid flags, I/O errors, etc), 2 to indicate a corrupt or\n"
+          "invalid input file, 3 for an internal consistency error (eg, bug) which\n"
+          "caused clzip to panic.\n"
+          "\nReport bugs to lzip-bug@nongnu.org\n"
+          "Clzip home page: http://www.nongnu.org/lzip/clzip.html\n" );
   }
 
 
-static void show_version()
+static void show_version( void )
   {
   printf( "%s %s\n", Program_name, PROGVERSION );
   printf( "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
-  printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n" );
-  printf( "This is free software: you are free to change and redistribute it.\n" );
-  printf( "There is NO WARRANTY, to the extent permitted by law.\n" );
+  printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
+          "This is free software: you are free to change and redistribute it.\n"
+          "There is NO WARRANTY, to the extent permitted by law.\n" );
   }
 
 
-static const char * format_num( long long num )
+static void show_header( const File_header header )
   {
   const char * const prefix[8] =
     { "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi" };
-  enum { buf_size = 16, factor = 1024 };
-  static char buf[buf_size];
-  const char *p = "";
-  int i;
+  enum { factor = 1024 };
+  const char * p = "";
+  const char * np = "  ";
+  unsigned num = Fh_get_dictionary_size( header ), i;
+  bool exact = ( num % factor == 0 );
 
-  for( i = 0; i < 8 && ( llabs( num ) > 9999 ||
-       ( llabs( num ) >= factor && num % factor == 0 ) ); ++i )
-    { num /= factor; p = prefix[i]; }
-  snprintf( buf, buf_size, "%lld %s", num, p );
-  return buf;
+  for( i = 0; i < 8 && ( num > 9999 || ( exact && num >= factor ) ); ++i )
+    { num /= factor; if( num % factor != 0 ) exact = false;
+      p = prefix[i]; np = ""; }
+  fprintf( stderr, "dictionary size %s%4u %sB.  ", np, num, p );
   }
 
 
-static long long getnum( const char * const ptr,
-                         const long long llimit, const long long ulimit )
+static unsigned long long getnum( const char * const ptr,
+                                  const unsigned long long llimit,
+                                  const unsigned long long ulimit )
   {
-  long long result;
-  char *tail;
+  unsigned long long result;
+  char * tail;
   errno = 0;
-  result = strtoll( ptr, &tail, 0 );
+  result = strtoull( ptr, &tail, 0 );
   if( tail == ptr )
     {
     show_error( "Bad or missing numerical argument.", 0, true );
@@ -192,8 +179,7 @@ static long long getnum( const char * const ptr,
   if( !errno && tail[0] )
     {
     int factor = ( tail[1] == 'i' ) ? 1024 : 1000;
-    int exponent = 0;
-    int i;
+    int exponent = 0, i;
     bool bad_multiplier = false;
     switch( tail[0] )
       {
@@ -218,7 +204,7 @@ static long long getnum( const char * const ptr,
       }
     for( i = 0; i < exponent; ++i )
       {
-      if( LLONG_MAX / factor >= llabs( result ) ) result *= factor;
+      if( ulimit / factor >= result ) result *= factor;
       else { errno = ERANGE; break; }
       }
     }
@@ -234,7 +220,7 @@ static long long getnum( const char * const ptr,
 
 static int get_dict_size( const char * const arg )
   {
-  char *tail;
+  char * tail;
   int bits = strtol( arg, &tail, 0 );
   if( bits >= min_dictionary_bits &&
       bits <= max_dictionary_bits && *tail == 0 )
@@ -243,59 +229,6 @@ static int get_dict_size( const char * const arg )
   }
 
 
-static int open_instream( const char * const name, struct stat * const in_statsp,
-                          const enum Mode program_mode, const int eindex,
-                          const bool recompress, const bool to_stdout )
-  {
-  int infd = -1;
-  if( program_mode == m_compress && !recompress && eindex >= 0 )
-    {
-    if( verbosity >= 0 )
-      fprintf( stderr, "%s: Input file `%s' already has `%s' suffix.\n",
-               program_name, name, known_extensions[eindex].from );
-    }
-  else
-    {
-    infd = open( name, O_RDONLY | o_binary );
-    if( infd < 0 )
-      {
-      if( verbosity >= 0 )
-        fprintf( stderr, "%s: Can't open input file `%s': %s.\n",
-                 program_name, name, strerror( errno ) );
-      }
-    else
-      {
-      const int i = fstat( infd, in_statsp );
-      const mode_t mode = in_statsp->st_mode;
-      const bool can_read = ( i == 0 &&
-                              ( S_ISBLK( mode ) || S_ISCHR( mode ) ||
-                                S_ISFIFO( mode ) || S_ISSOCK( mode ) ) );
-      if( i != 0 || ( !S_ISREG( mode ) && ( !to_stdout || !can_read ) ) )
-        {
-        if( verbosity >= 0 )
-          fprintf( stderr, "%s: Input file `%s' is not a regular file%s.\n",
-                   program_name, name,
-                   ( can_read && !to_stdout ) ?
-                   " and `--stdout' was not specified" : "" );
-        close( infd );
-        infd = -1;
-        }
-      }
-    }
-  return infd;
-  }
-
-
-static void set_c_outname( const char * const name, const bool multifile )
-  {
-  output_filename = resize_buffer( output_filename, strlen( name ) + 5 +
-                                   strlen( known_extensions[0].from ) + 1 );
-  strcpy( output_filename, name );
-  if( multifile ) strcat( output_filename, "00001" );
-  strcat( output_filename, known_extensions[0].from );
-  }
-
-
 static int extension_index( const char * const name )
   {
   int i;
@@ -310,6 +243,74 @@ static int extension_index( const char * const name )
   }
 
 
+static int open_instream( const char * const name, struct stat * const in_statsp,
+                          const enum Mode program_mode, const int eindex,
+                          const bool recompress, const bool to_stdout )
+  {
+  int infd = -1;
+  if( program_mode == m_compress && !recompress && eindex >= 0 )
+    {
+    if( verbosity >= 0 )
+      fprintf( stderr, "%s: Input file '%s' already has '%s' suffix.\n",
+               program_name, name, known_extensions[eindex].from );
+    }
+  else
+    {
+    infd = open( name, O_RDONLY | o_binary );
+    if( infd < 0 )
+      {
+      if( verbosity >= 0 )
+        fprintf( stderr, "%s: Can't open input file '%s': %s.\n",
+                 program_name, name, strerror( errno ) );
+      }
+    else
+      {
+      const int i = fstat( infd, in_statsp );
+      const mode_t mode = in_statsp->st_mode;
+      const bool can_read = ( i == 0 &&
+                              ( S_ISBLK( mode ) || S_ISCHR( mode ) ||
+                                S_ISFIFO( mode ) || S_ISSOCK( mode ) ) );
+      const bool no_ofile = to_stdout || program_mode == m_test;
+      if( i != 0 || ( !S_ISREG( mode ) && ( !can_read || !no_ofile ) ) )
+        {
+        if( verbosity >= 0 )
+          fprintf( stderr, "%s: Input file '%s' is not a regular file%s.\n",
+                   program_name, name,
+                   ( can_read && !no_ofile ) ?
+                   " and '--stdout' was not specified" : "" );
+        close( infd );
+        infd = -1;
+        }
+      }
+    }
+  return infd;
+  }
+
+
+/* assure at least a minimum size for buffer 'buf' */
+static void * resize_buffer( void * buf, const int min_size )
+  {
+  if( buf ) buf = realloc( buf, min_size );
+  else buf = malloc( min_size );
+  if( !buf )
+    {
+    show_error( "Not enough memory.", 0, false );
+    cleanup_and_fail( 1 );
+    }
+  return buf;
+  }
+
+
+static void set_c_outname( const char * const name, const bool multifile )
+  {
+  output_filename = resize_buffer( output_filename, strlen( name ) + 5 +
+                                   strlen( known_extensions[0].from ) + 1 );
+  strcpy( output_filename, name );
+  if( multifile ) strcat( output_filename, "00001" );
+  strcat( output_filename, known_extensions[0].from );
+  }
+
+
 static void set_d_outname( const char * const name, const int i )
   {
   if( i >= 0 )
@@ -329,7 +330,7 @@ static void set_d_outname( const char * const name, const int i )
   strcpy( output_filename, name );
   strcat( output_filename, ".out" );
   if( verbosity >= 1 )
-    fprintf( stderr, "%s: Can't guess original name for `%s' -- using `%s'.\n",
+    fprintf( stderr, "%s: Can't guess original name for '%s' -- using '%s'.\n",
              program_name, name, output_filename );
   }
 
@@ -343,10 +344,10 @@ static bool open_outstream( const bool force )
   if( outfd < 0 && verbosity >= 0 )
     {
     if( errno == EEXIST )
-      fprintf( stderr, "%s: Output file %s already exists, skipping.\n",
+      fprintf( stderr, "%s: Output file '%s' already exists, skipping.\n",
                program_name, output_filename );
     else
-      fprintf( stderr, "%s: Can't create output file `%s': %s.\n",
+      fprintf( stderr, "%s: Can't create output file '%s': %s.\n",
                program_name, output_filename, strerror( errno ) );
     }
   return ( outfd >= 0 );
@@ -376,7 +377,7 @@ void cleanup_and_fail( const int retval )
     {
     delete_output_on_interrupt = false;
     if( verbosity >= 0 )
-      fprintf( stderr, "%s: Deleting output file `%s', if it exists.\n",
+      fprintf( stderr, "%s: Deleting output file '%s', if it exists.\n",
                program_name, output_filename );
     if( outfd >= 0 ) { close( outfd ); outfd = -1; }
     if( remove( output_filename ) != 0 && errno != ENOENT )
@@ -389,37 +390,32 @@ void cleanup_and_fail( const int retval )
      /* Set permissions, owner and times. */
 static void close_and_set_permissions( const struct stat * const in_statsp )
   {
-  bool error = false;
+  bool warning = false;
   if( in_statsp )
     {
+    /* fchown will in many cases return with EPERM, which can be safely ignored. */
     if( ( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) != 0 &&
           errno != EPERM ) ||
-        fchmod( outfd, in_statsp->st_mode ) != 0 )
-      error = true;
-    /* fchown will in many cases return with EPERM, which can be safely ignored. */
+        fchmod( outfd, in_statsp->st_mode ) != 0 ) warning = true;
     }
-  if( close( outfd ) == 0 ) outfd = -1;
-  else cleanup_and_fail( 1 );
+  if( close( outfd ) != 0 ) cleanup_and_fail( 1 );
+  outfd = -1;
   delete_output_on_interrupt = false;
-  if( !in_statsp ) return;
-  if( !error )
+  if( in_statsp )
     {
     struct utimbuf t;
     t.actime = in_statsp->st_atime;
     t.modtime = in_statsp->st_mtime;
-    if( utime( output_filename, &t ) != 0 ) error = true;
+    if( utime( output_filename, &t ) != 0 ) warning = true;
     }
-  if( error )
-    {
+  if( warning && verbosity >= 1 )
     show_error( "Can't change output file attributes.", 0, false );
-    cleanup_and_fail( 1 );
-    }
   }
 
 
-static bool next_filename()
+static bool next_filename( void )
   {
-  const unsigned int len = strlen( known_extensions[0].from );
+  const unsigned len = strlen( known_extensions[0].from );
   int i, j;
 
   if( strlen( output_filename ) >= len + 5 )		/* "*00001.lz" */
@@ -432,50 +428,63 @@ static bool next_filename()
   }
 
 
-static int compress( const long long member_size, const long long volume_size,
+static int compress( const unsigned long long member_size,
+                     const unsigned long long volume_size,
                      const struct Lzma_options * const encoder_options,
                      const int infd, struct Pretty_print * const pp,
                      const struct stat * const in_statsp )
   {
-  long long in_size = 0, out_size = 0, partial_volume_size = 0;
+  unsigned long long in_size = 0, out_size = 0, partial_volume_size = 0;
   int retval = 0;
-  File_header header;
   struct Matchfinder matchfinder;
+  File_header header;
+  Fh_set_magic( header );
 
   if( verbosity >= 1 ) Pp_show_msg( pp, 0 );
-  Fh_set_magic( header );
   if( !Fh_set_dictionary_size( header, encoder_options->dictionary_size ) ||
       encoder_options->match_len_limit < min_match_len_limit ||
       encoder_options->match_len_limit > max_match_len )
     internal_error( "invalid argument to encoder" );
 
-  Mf_init( &matchfinder, Fh_get_dictionary_size( header ),
-           encoder_options->match_len_limit, infd );
-  Fh_set_dictionary_size( header, Mf_dictionary_size( &matchfinder ) );
+  if( !Mf_init( &matchfinder, Fh_get_dictionary_size( header ),
+                encoder_options->match_len_limit, infd ) )
+    {
+    Pp_show_msg( pp, "Not enough memory. Try a smaller dictionary size" );
+    return 1;
+    }
+  Fh_set_dictionary_size( header, matchfinder.dictionary_size );
 
-  while( true )		/* encode one member per iteration */
+  while( true )			/* encode one member per iteration */
     {
     struct LZ_encoder encoder;
-    const long long size =
-      min( member_size, volume_size - partial_volume_size );
-    LZe_init( &encoder, &matchfinder, header, outfd );
+    const unsigned long long size = ( volume_size > 0 ) ?
+      min( member_size, volume_size - partial_volume_size ) : member_size;
+    if( !LZe_init( &encoder, &matchfinder, header, outfd ) )
+      {
+      show_error( "Not enough memory. Try a smaller dictionary size.", 0, false );
+      cleanup_and_fail( 1 );
+      }
+    show_progress( in_size, &matchfinder, pp, in_statsp );	/* init */
     if( !LZe_encode_member( &encoder, size ) )
       { Pp_show_msg( pp, "Encoder error" ); retval = 1; break; }
     in_size += Mf_data_position( &matchfinder );
-    out_size += LZe_member_position( &encoder );
-    partial_volume_size += LZe_member_position( &encoder );
+    out_size += Re_member_position( &encoder.renc );
     LZe_free( &encoder );
     if( Mf_finished( &matchfinder ) ) break;
-    if( partial_volume_size >= volume_size - min_dictionary_size )
+    if( volume_size > 0 )
       {
-      partial_volume_size = 0;
-      if( delete_output_on_interrupt )
+      partial_volume_size += Re_member_position( &encoder.renc );
+      if( partial_volume_size >= volume_size - min_dictionary_size )
         {
-        close_and_set_permissions( in_statsp );
-        if( !next_filename() )
-          { Pp_show_msg( pp, "Too many volume files" ); retval = 1; break; }
-        if( !open_outstream( true ) ) { retval = 1; break; }
-        delete_output_on_interrupt = true;
+        partial_volume_size = 0;
+        if( delete_output_on_interrupt )
+          {
+          close_and_set_permissions( in_statsp );
+          if( !next_filename() )
+            { Pp_show_msg( pp, "Too many volume files" ); retval = 1; break; }
+          if( !open_outstream( true ) ) { retval = 1; break; }
+          delete_output_on_interrupt = true;
+          }
         }
       }
     Mf_reset( &matchfinder );
@@ -483,11 +492,11 @@ static int compress( const long long member_size, const long long volume_size,
 
   if( retval == 0 && verbosity >= 1 )
     {
-    if( in_size <= 0 || out_size <= 0 )
-      fprintf( stderr, "No data compressed.\n" );
+    if( in_size == 0 || out_size == 0 )
+      fprintf( stderr, " no data compressed.\n" );
     else
       fprintf( stderr, "%6.3f:1, %6.3f bits/byte, "
-                       "%5.2f%% saved, %lld in, %lld out.\n",
+                       "%5.2f%% saved, %llu in, %llu out.\n",
                (double)in_size / out_size,
                ( 8.0 * out_size ) / in_size,
                100.0 * ( 1.0 - ( (double)out_size / in_size ) ),
@@ -501,23 +510,28 @@ static int compress( const long long member_size, const long long volume_size,
 static int decompress( const int infd, struct Pretty_print * const pp,
                        const bool testing )
   {
-  long long partial_file_pos = 0;
+  unsigned long long partial_file_pos = 0;
   struct Range_decoder rdec;
-  int retval = 0, i, result;
+  int retval = 0;
   bool first_member;
-  Rd_init( &rdec, infd );
-
-  for( first_member = true; ; first_member = false, Pp_reset( pp ) )
+  if( !Rd_init( &rdec, infd ) )
     {
+    show_error( "Not enough memory. Find a machine with more memory.", 0, false );
+    cleanup_and_fail( 1 );
+    }
+
+  for( first_member = true; ; first_member = false )
+    {
+    int result;
     File_header header;
     struct LZ_decoder decoder;
     Rd_reset_member_position( &rdec );
-    for( i = 0; i < Fh_size; ++i )
-      header[i] = Rd_get_byte( &rdec );
-    if( Rd_finished( &rdec ) )		/* End Of File */
+    Rd_read_data( &rdec, header, Fh_size );
+    if( Rd_finished( &rdec ) )			/* End Of File */
       {
       if( first_member )
-        { Pp_show_msg( pp, "Error reading member header" ); retval = 1; }
+        { Pp_show_msg( pp, "File ends unexpectedly at member header" );
+          retval = 2; }
       break;
       }
     if( !Fh_verify_magic( header ) )
@@ -540,15 +554,13 @@ static int decompress( const int infd, struct Pretty_print * const pp,
         retval = 2; break; }
 
     if( verbosity >= 2 || ( verbosity == 1 && first_member ) )
-      {
-      Pp_show_msg( pp, 0 );
-      if( verbosity >= 2 )
-        fprintf( stderr, "version %d, dictionary size %7sB.  ",
-                 Fh_version( header ),
-                 format_num( Fh_get_dictionary_size( header ) ) );
-      }
-    LZd_init( &decoder, header, &rdec, outfd );
+      { Pp_show_msg( pp, 0 ); if( verbosity >= 3 ) show_header( header ); }
 
+    if( !LZd_init( &decoder, header, &rdec, outfd ) )
+      {
+      show_error( "Not enough memory. Find a machine with more memory.", 0, false );
+      cleanup_and_fail( 1 );
+      }
     result = LZd_decode_member( &decoder, pp );
     partial_file_pos += Rd_member_position( &rdec );
     LZd_free( &decoder );
@@ -558,34 +570,32 @@ static int decompress( const int infd, struct Pretty_print * const pp,
         {
         Pp_show_msg( pp, 0 );
         if( result == 2 )
-          fprintf( stderr, "File ends unexpectedly at pos %lld\n",
+          fprintf( stderr, "File ends unexpectedly at pos %llu\n",
                    partial_file_pos );
         else
-          fprintf( stderr, "Decoder error at pos %lld\n", partial_file_pos );
+          fprintf( stderr, "Decoder error at pos %llu\n", partial_file_pos );
         }
       retval = 2; break;
       }
     if( verbosity >= 2 )
-      { if( testing ) fprintf( stderr, "ok\n" );
-        else fprintf( stderr, "done\n" ); }
+      { fprintf( stderr, testing ? "ok\n" : "done\n" ); Pp_reset( pp ); }
     }
   Rd_free( &rdec );
   if( verbosity == 1 && retval == 0 )
-    { if( testing ) fprintf( stderr, "ok\n" );
-      else fprintf( stderr, "done\n" ); }
+    fprintf( stderr, testing ? "ok\n" : "done\n" );
   return retval;
   }
 
 
 void signal_handler( int sig )
   {
-  sig = 0;			/* keep compiler happy */
+  if( sig ) {}				/* keep compiler happy */
   show_error( "Control-C or similar caught, quitting.", 0, false );
   cleanup_and_fail( 1 );
   }
 
 
-static void set_signals()
+static void set_signals( void )
   {
   signal( SIGHUP, signal_handler );
   signal( SIGINT, signal_handler );
@@ -594,45 +604,26 @@ static void set_signals()
 
 
 void Pp_init( struct Pretty_print * const pp, const char * const filenames[],
-              const int num_filenames, const int v )
+              const int num_filenames )
   {
-  unsigned int stdin_name_len;
+  unsigned stdin_name_len;
   int i;
-  pp->name_ = 0;
+  pp->name = 0;
   pp->stdin_name = "(stdin)";
   pp->longest_name = 0;
-  pp->verbosity = v;
   pp->first_post = false;
   stdin_name_len = strlen( pp->stdin_name );
 
   for( i = 0; i < num_filenames; ++i )
     {
     const char * const s = filenames[i];
-    const int len = ( !strcmp( s, "-" ) ? stdin_name_len : strlen( s ) );
+    const int len = ( (strcmp( s, "-" ) == 0) ? stdin_name_len : strlen( s ) );
     if( len > pp->longest_name ) pp->longest_name = len;
     }
   if( pp->longest_name == 0 ) pp->longest_name = stdin_name_len;
   }
 
 
-void Pp_show_msg( struct Pretty_print * const pp, const char * const msg )
-  {
-  if( verbosity >= 0 )
-    {
-    if( pp->first_post )
-      {
-      int i, len;
-      pp->first_post = false;
-      fprintf( stderr, "  %s: ", pp->name_ );
-      len = pp->longest_name - strlen( pp->name_ );
-      for( i = 0; i < len; ++i ) fprintf( stderr, " " );
-      if( !msg ) fflush( stderr );
-      }
-    if( msg ) fprintf( stderr, "%s.\n", msg );
-    }
-  }
-
-
 void show_error( const char * const msg, const int errcode, const bool help )
   {
   if( verbosity >= 0 )
@@ -643,8 +634,8 @@ void show_error( const char * const msg, const int errcode, const bool help )
       if( errcode > 0 ) fprintf( stderr, ": %s", strerror( errcode ) );
       fprintf( stderr, "\n" );
       }
-    if( help && invocation_name && invocation_name[0] )
-      fprintf( stderr, "Try `%s --help' for more information.\n",
+    if( help )
+      fprintf( stderr, "Try '%s --help' for more information.\n",
                invocation_name );
     }
   }
@@ -658,6 +649,34 @@ void internal_error( const char * const msg )
   }
 
 
+void show_progress( const unsigned long long partial_size,
+                    const struct Matchfinder * const m,
+                    struct Pretty_print * const p,
+                    const struct stat * const in_statsp )
+  {
+  static unsigned long long cfile_size = 0;	/* file_size / 100 */
+  static unsigned long long psize = 0;
+  static const struct Matchfinder * mf = 0;
+  static struct Pretty_print * pp = 0;
+
+  if( m )					/* initialize static vars */
+    {
+    psize = partial_size; mf = m; pp = p;
+    cfile_size = ( in_statsp && S_ISREG( in_statsp->st_mode ) ) ?
+                 in_statsp->st_size / 100 : 0;
+    return;
+    }
+  if( mf && pp )
+    {
+    const unsigned long long pos = psize + Mf_data_position( mf );
+    if( cfile_size > 0 )
+      fprintf( stderr, "%4llu%%", pos / cfile_size );
+    fprintf( stderr, "  %.1f MB\r", pos / 1000000.0 );
+    Pp_reset( pp ); Pp_show_msg( pp, 0 );	/* restore cursor position */
+    }
+  }
+
+
 int main( const int argc, const char * const argv[] )
   {
   /* Mapping from gzip/bzip2 style 1..9 compression modes
@@ -675,8 +694,10 @@ int main( const int argc, const char * const argv[] )
     { 3 << 23, 132 },		/* -8 */
     { 1 << 25, 273 } };		/* -9 */
   struct Lzma_options encoder_options = option_mapping[6];  /* default = "-6" */
-  long long member_size = LLONG_MAX;
-  long long volume_size = LLONG_MAX;
+  const unsigned long long max_member_size = 0x0100000000000000ULL;
+  const unsigned long long max_volume_size = 0x4000000000000000ULL;
+  unsigned long long member_size = max_member_size;
+  unsigned long long volume_size = 0;
   const char * input_filename = "";
   const char * default_output_filename = "";
   const char ** filenames = 0;
@@ -708,12 +729,12 @@ int main( const int argc, const char * const argv[] )
     { 'b', "member-size",     ap_yes },
     { 'c', "stdout",          ap_no  },
     { 'd', "decompress",      ap_no  },
-    { 'e', "extreme",         ap_no  },
     { 'f', "force",           ap_no  },
     { 'F', "recompress",      ap_no  },
     { 'h', "help",            ap_no  },
     { 'k', "keep",            ap_no  },
     { 'm', "match-length",    ap_yes },
+    { 'n', "threads",         ap_yes },
     { 'o', "output",          ap_yes },
     { 'q', "quiet",           ap_no  },
     { 's', "dictionary-size", ap_yes },
@@ -727,6 +748,7 @@ int main( const int argc, const char * const argv[] )
 
   invocation_name = argv[0];
   CRC32_init();
+
   if( !ap_init( &parser, argc, argv, options, 0 ) )
     { show_error( "Memory exhausted.", 0, false ); return 1; }
   if( ap_error( &parser ) )				/* bad option */
@@ -736,28 +758,28 @@ int main( const int argc, const char * const argv[] )
     {
     const int code = ap_code( &parser, argind );
     const char * const arg = ap_argument( &parser, argind );
-    if( !code ) break;				/* no more options */
+    if( !code ) break;					/* no more options */
     switch( code )
       {
       case '0':
       case '1': case '2': case '3': case '4':
       case '5': case '6': case '7': case '8': case '9':
                 encoder_options = option_mapping[code-'0']; break;
-      case 'b': member_size = getnum( arg, 100000, LLONG_MAX / 2 ); break;
+      case 'b': member_size = getnum( arg, 100000, max_member_size ); break;
       case 'c': to_stdout = true; break;
       case 'd': program_mode = m_decompress; break;
-      case 'e': break;				/* ignored by now */
       case 'f': force = true; break;
       case 'F': recompress = true; break;
       case 'h': show_help(); return 0;
       case 'k': keep_input_files = true; break;
       case 'm': encoder_options.match_len_limit =
                   getnum( arg, min_match_len_limit, max_match_len ); break;
+      case 'n': break;
       case 'o': default_output_filename = arg; break;
       case 'q': verbosity = -1; break;
       case 's': encoder_options.dictionary_size = get_dict_size( arg );
                 break;
-      case 'S': volume_size = getnum( arg, 100000, LLONG_MAX / 2 ); break;
+      case 'S': volume_size = getnum( arg, 100000, max_volume_size ); break;
       case 't': program_mode = m_test; break;
       case 'v': if( verbosity < 4 ) ++verbosity; break;
       case 'V': show_version(); return 0;
@@ -766,30 +788,10 @@ int main( const int argc, const char * const argv[] )
     } /* end process options */
 
 #if defined(__MSVCRT__) || defined(__OS2__)
-  _setmode( STDIN_FILENO, O_BINARY );
-  _setmode( STDOUT_FILENO, O_BINARY );
+  setmode( STDIN_FILENO, O_BINARY );
+  setmode( STDOUT_FILENO, O_BINARY );
 #endif
 
-  for( ; argind < ap_arguments( &parser ); ++argind )
-    {
-    if( strcmp( ap_argument( &parser, argind ), "-" ) )
-      filenames_given = true;
-    ++num_filenames;
-    filenames = resize_buffer( filenames, num_filenames * sizeof (char *) );
-    filenames[num_filenames-1] = ap_argument( &parser, argind );
-    }
-
-  if( num_filenames == 0 )
-    {
-    ++num_filenames;
-    filenames = resize_buffer( filenames, sizeof (char *) );
-    filenames[num_filenames-1] = "-";
-    }
-  if( !to_stdout && program_mode != m_test &&
-      ( filenames_given || default_output_filename[0] ) )
-    set_signals();
-
-  Pp_init( &pp, filenames, num_filenames, verbosity );
   if( program_mode == m_test )
     outfd = -1;
   else if( program_mode == m_compress )
@@ -798,6 +800,22 @@ int main( const int argc, const char * const argv[] )
     Prob_prices_init();
     }
 
+  num_filenames = max( 1, ap_arguments( &parser ) - argind );
+  filenames = resize_buffer( filenames, num_filenames * sizeof filenames[0] );
+  filenames[0] = "-";
+
+  for( i = 0; argind + i < ap_arguments( &parser ); ++i )
+    {
+    filenames[i] = ap_argument( &parser, argind + i );
+    if( strcmp( filenames[i], "-" ) != 0 ) filenames_given = true;
+    }
+
+  if( !to_stdout && program_mode != m_test &&
+      ( filenames_given || default_output_filename[0] ) )
+    set_signals();
+
+  Pp_init( &pp, filenames, num_filenames );
+
   output_filename = resize_buffer( output_filename, 1 );
   for( i = 0; i < num_filenames; ++i )
     {
@@ -806,7 +824,7 @@ int main( const int argc, const char * const argv[] )
     const struct stat * in_statsp;
     output_filename[0] = 0;
 
-    if( !filenames[i][0] || !strcmp( filenames[i], "-" ) )
+    if( !filenames[i][0] || strcmp( filenames[i], "-" ) == 0 )
       {
       input_filename = "";
       infd = STDIN_FILENO;
@@ -817,17 +835,17 @@ int main( const int argc, const char * const argv[] )
         else
           {
           if( program_mode == m_compress )
-            set_c_outname( default_output_filename, volume_size != LLONG_MAX );
+            set_c_outname( default_output_filename, volume_size > 0 );
           else
             {
             output_filename = resize_buffer( output_filename,
                                              strlen( default_output_filename ) + 1 );
             strcpy( output_filename, default_output_filename );
             }
-          outfd_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
+          outfd_mode = all_rw;
           if( !open_outstream( force ) )
             {
-            if( outfd == -1 && retval < 1 ) retval = 1;
+            if( retval < 1 ) retval = 1;
             close( infd ); infd = -1;
             continue;
             }
@@ -847,12 +865,12 @@ int main( const int argc, const char * const argv[] )
         else
           {
           if( program_mode == m_compress )
-            set_c_outname( input_filename, volume_size != LLONG_MAX );
+            set_c_outname( input_filename, volume_size > 0 );
           else set_d_outname( input_filename, eindex );
-          outfd_mode = S_IRUSR | S_IWUSR;
+          outfd_mode = usr_rw;
           if( !open_outstream( force ) )
             {
-            if( outfd == -1 && retval < 1 ) retval = 1;
+            if( retval < 1 ) retval = 1;
             close( infd ); infd = -1;
             continue;
             }
